#ifndef RF24_UTILITY_INCLUDES_H_
#define RF24_UTILITY_INCLUDES_H_

#define RF24_PIGPIO

#include <cstring> // memcpy() used in RF24.cpp
#include "pigpio/RF24_arch_config.h"
#include "pigpio/interrupt.h"

#endif
